# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2020 Red Hat, Inc., see CONTRIBUTORS.


import fingertip.machine
from fingertip.util import log, reflink


@fingertip.transient
def main(what=None):
    if what in ('setup', 'unmount', 'cleanup'):
        return globals()[what]()
    log.error('usage: ')
    log.error('    fingertip filesystem setup')
    log.error('    fingertip filesystem unmount')
    log.error('    fingertip filesystem cleanup')
    raise SystemExit()


def setup():
    reflink.storage_setup_wizard()


def unmount():
    reflink.storage_unmount()


def cleanup():
    reflink.storage_destroy()



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import datetime
import functools
import inspect
import os

import cloudpickle

import fingertip.exec
from fingertip import step_loader, expiration
from fingertip.util import hooks, lock, log, path, reflink, temp


def transient(func=None, when='always'):
    if func is None:  # no parameters, just @transient
        return functools.partial(transient, when=when)

    func.transient = when  # can be a callable!
    return func


def supply_last_step_if_requested(func, fingertip_last_step):
    if 'fingertip_last_step' in inspect.signature(func).parameters:
        return functools.partial(func, fingertip_last_step=fingertip_last_step)
    return func


class Machine:
    def __init__(self, backend_name, sealed=True, expire_in='7d'):
        self.hooks = hooks.HookManager()
        os.makedirs(path.MACHINES, exist_ok=True)
        self.path = temp.disappearing_dir(path.MACHINES)
        self._parent_path = path.MACHINES
        # States: loaded -> spun_up -> spun_down -> saved/dropped
        self._state = 'spun_down'
        self._transient = False
        self._up_counter = 0
        self.sealed = sealed
        self.expiration = expiration.Expiration(expire_in)
        self.backend = backend_name
        self.log = log.Sublogger(f'plugins.backend.{backend_name}',
                                 os.path.join(self.path, 'log.txt'))
        self.log.debug(f'created {backend_name}')
        self.hooks.clone.append(
            lambda to: reflink.auto(os.path.join(self.path, 'log.txt'),
                                    os.path.join(to, 'log.txt')))

    def __call__(self, *args, **kwargs):  # a convenience method
        return fingertip.exec.nice_exec(self, *args, **kwargs)

    def transient(self):
        self._transient = True
        return self

    def __enter__(self):
        log.debug(f'state={self._state}')
        assert (self._state == 'loaded' and not self._up_counter or
                self._state == 'spun_up' and self._up_counter)
        if not self._up_counter:
            assert self._state == 'loaded'
            self.hooks.up()
            self._state = 'spun_up'
        self._up_counter += 1
        return self

    def __exit__(self, exc_type, *_):
        assert self._state == 'spun_up'
        self._up_counter -= 1
        if not self._up_counter:
            if not self._transient:
                # the machine needs to be spun down, will be finalized later
                self.hooks.down.in_reverse()
                self._state = 'spun_down'
            else:
                # the machine can be dropped and finalized, fast and dirty
                self.hooks.drop.in_reverse()
                self._state = 'dropped'
                if not exc_type:
                    self._finalize()

    def _finalize(self, link_as=None, name_hint=None):
        log.debug(f'finalize hint={name_hint} link_as={link_as} {self._state}')
        if link_as and self._state == 'spun_down':
            self.hooks.save.in_reverse()
            temp_path = self.path
            self.path = temp.unique_dir(self._parent_path, hint=name_hint)
            log.debug(f'saving to temp {temp_path}')
            self._state = 'saving'
            self.expiration.depend_on_loaded_python_modules()
            self.log.finalize()
            with open(os.path.join(temp_path, 'machine.clpickle'), 'wb') as f:
                cloudpickle.dump(self, f)
            log.debug(f'moving {temp_path} to {self.path}')
            os.rename(temp_path, self.path)
            self._state == 'saved'
            link_this = self.path
        else:
            self.log.finalize()
            assert self._state in ('spun_down', 'loaded', 'dropped')
            log.info(f'discarding {self.path}')
            # TODO: track whether the step will be transient-last and reflink?
            with open(os.path.join(self.path, 'log.txt')) as f:
                self.log_contents = f.read()
            temp.remove(self.path)
            link_this = self._parent_path
            self._state = 'dropped'
        if (link_this and link_as and
                os.path.realpath(link_as) != os.path.realpath(link_this)):
            log.debug(f'linking {link_this} to {link_as}')
            if os.path.lexists(link_as):
                if os.path.exists(link_as) and not needs_a_rebuild(link_as):
                    log.critical(f'Refusing to overwrite fresh {link_as}')
                    raise RuntimeError(f'Not overriding fresh {link_as}')
                os.unlink(link_as)
            os.symlink(link_this, link_as)
            return link_as

    def apply(self, step, *args, fingertip_last_step=False, **kwargs):
        func, tag = step_loader.func_and_autotag(step, *args, **kwargs)
        log.debug(f'apply {self.path} {step} {func} {args} {kwargs}')
        if self._state == 'spun_up':
            log.debug(f'applying to unclean')
            func = supply_last_step_if_requested(func, fingertip_last_step)
            return func(self, *args, **kwargs)
        elif self._state == 'loaded':
            log.debug(f'applying to clean')
            return self._cache_aware_apply(step, tag, func, args, kwargs,
                                           fingertip_last_step)
        else:
            log.critical(f'apply to state={self._state}')
            raise RuntimeError(f'State machine error, apply to {self._state}')

    def _cache_aware_apply(self, step, tag, func, args, kwargs, last_step):
        assert self._state == 'loaded'

        transient_hint = func.transient if hasattr(func, 'transient') else None
        if callable(transient_hint):
            transient_hint = supply_last_step_if_requested(transient_hint,
                                                           last_step)
            transient_hint = transient_hint(self, *args, **kwargs)

        return_as_transient = self._transient
        exec_as_transient = (
            transient_hint in ('always', True) or
            transient_hint == 'last' and last_step
        )
        log.debug(f'transient: {transient_hint}')
        log.debug(f'exec_as_transient: {exec_as_transient}')
        log.debug(f'return_as_transient: {return_as_transient}')
        self._transient = exec_as_transient

        # Could there already be a cached result?
        log.debug(f'PATH {self.path} {tag}')
        new_mpath = os.path.join(self._parent_path, tag)

        lock_path = os.path.join(self._parent_path, '.' + tag + '-lock')
        do_lock = not self._transient
        if do_lock:
            log.info(f'acquiring lock for {tag}...')
        prev_log_name = self.log.name
        self.log.finalize()
        with lock.MaybeLock(lock_path, lock=do_lock):
            if (os.path.exists(new_mpath) and not needs_a_rebuild(new_mpath)
                    and not exec_as_transient):
                # sweet, scratch this instance, fast-forward to cached result
                log.info(f'reusing {step} @ {new_mpath}')
                self._finalize()
                clone_from_path = new_mpath
            else:
                # loaded, not spun up, step not cached: perform step, cache
                log.info(f'applying (and, possibly, caching) {tag}')
                self.log = log.Sublogger('plugins.' + tag.split(':', 1)[0],
                                         os.path.join(self.path, 'log.txt'))
                func = supply_last_step_if_requested(func, last_step)
                m = func(self, *args, **kwargs)
                if m:
                    if m._transient and transient_hint == 'last' and last_step:
                        assert m._state == 'dropped'
                        # transient-when-last step returned m
                        # just in case it's not the last, but it was.
                        # m is dropped already, only log contents is preserved.
                        fname = f'{datetime.datetime.utcnow().isoformat()}.txt'
                        t = path.logs(fname, makedirs=True)
                        with open(t, 'w') as f:
                            f.write(m.log_contents)
                        return t
                    assert not m._transient, 'transient step returned a value'
                    m._finalize(link_as=new_mpath, name_hint=tag)
                    clone_from_path = new_mpath
                    log.info(f'successfully applied and saved {tag}')
                else:  # transient step, either had hints or just returned None
                    clone_from_path = self._parent_path
                    log.info(f'successfully applied and dropped {tag}')
        if last_step:
            return os.path.join(clone_from_path, 'log.txt')
        m = clone_and_load(clone_from_path)
        m.log = log.Sublogger(prev_log_name, os.path.join(m.path, 'log.txt'))
        m._transient = return_as_transient
        return m


def _load_from_path(data_dir_path):
    log.debug(f'load from {data_dir_path}')
    with open(os.path.join(data_dir_path, 'machine.clpickle'), 'rb') as f:
        m = cloudpickle.load(f)
    assert m._state == 'saving'
    m._state = 'loading'
    m.log = log.Sublogger('<unknown>')
    assert m.path == data_dir_path
    assert m._parent_path == os.path.realpath(os.path.dirname(data_dir_path))
    m.hooks.load()
    m._state = 'loaded'
    return m


def clone_and_load(from_path, name_hint=None):
    log.debug(f'clone {from_path}')
    temp_path = temp.disappearing_dir(from_path, hint=name_hint)
    log.debug(f'temp = {temp_path}')
    os.makedirs(temp_path, exist_ok=True)
    with open(os.path.join(from_path, 'machine.clpickle'), 'rb') as f:
        m = cloudpickle.load(f)
    m.log = log.Sublogger('<cloning>')
    m.hooks.clone(temp_path)
    m._parent_path = os.path.realpath(from_path)
    m.path = temp_path
    with open(os.path.join(m.path, 'machine.clpickle'), 'wb') as f:
        cloudpickle.dump(m, f)
    return _load_from_path(temp_path)


def build(first_step, *args, fingertip_last_step=False, **kwargs):
    func, tag = step_loader.func_and_autotag(first_step, *args, **kwargs)

    # Could there already be a cached result?
    mpath = path.machines(tag)
    lock_path = path.machines('.' + tag + '-lock')
    log.info(f'acquiring lock for {tag}...')

    transient_hint = func.transient if hasattr(func, 'transient') else None
    if callable(transient_hint):
        transient_hint = supply_last_step_if_requested(transient_hint,
                                                       fingertip_last_step)
        transient_hint = transient_hint(*args, **kwargs)
    transient = (
        transient_hint in ('always', True) or
        transient_hint == 'last' and fingertip_last_step
    )

    with lock.MaybeLock(lock_path, lock=not transient):
        if not os.path.exists(mpath) or needs_a_rebuild(mpath):
            log.info(f'building {tag}...')
            func = supply_last_step_if_requested(func, fingertip_last_step)
            first = func(*args, **kwargs)

            if first is None:
                assert transient, 'first step returned None'
                return

            if transient:
                log.info(f'succesfully built and discarded {tag}')
                first._finalize()  # discard (not fast-dropped though)

                if transient_hint == 'last' and fingertip_last_step:
                    fname = f'{datetime.datetime.utcnow().isoformat()}.txt'
                    t = path.logs(fname, makedirs=True)
                    with open(t, 'w') as f:
                        f.write(first.log_contents)
                    return t
            else:
                log.info(f'succesfully built and saved {tag}')
                first._finalize(link_as=mpath, name_hint=tag)

    if fingertip_last_step:
        return os.path.join(mpath, 'log.txt')
    m = clone_and_load(mpath)
    m.log = log.Sublogger('<just built>', os.path.join(m.path, 'log.txt'))
    return m


OFFLINE = os.getenv('FINGERTIP_OFFLINE', '0') != '0'


def needs_a_rebuild(mpath, by=None):
    with open(os.path.join(mpath, 'machine.clpickle'), 'rb') as f:
        m = cloudpickle.load(f)
    if not m.expiration.files_have_not_changed():
        return True
    expired = m.expiration.is_expired(by)
    if expired:
        log.debug(f'{mpath} has expired at {m.expiration.pretty()}')
    else:
        log.debug(f'{mpath} is valid until {m.expiration.pretty()}')
    if OFFLINE and expired:
        log.warning(f'{mpath} expired at {m.expiration.pretty()}, '
                    'but offline mode is enabled, so, reusing it')
    return expired and not OFFLINE



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import fingertip


@fingertip.transient
def main(m, next_plugin, *args, **kwargs):
    with m:
        m.apply(next_plugin, *args, **kwargs)



"""
Helper functions for fingertep: inter-process locking.
"""

import fasteners


class MaybeLock:
    def __init__(self, path, lock=True):
        self._lock = lock and fasteners.process_lock.InterProcessLock(path)

    def __enter__(self):
        if self._lock:
            self._lock.acquire()
        return self

    def __exit__(self, *_):
        if self._lock:
            self._lock.release()



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

"""
Helper functions for fingertip: working with self-destructing tempfiles.
"""

# TODO: replace with python3-tempdir

import atexit
import os
import signal
import shutil
import sys
import tempfile

from fingertip.util import log


AUTOREMOVE_PREFIX = 'tmp-fingertip.'


# TODO: hacky and unclean
def terminate_child(num, frame):
    print('caught SIGTERM, cleaning up...')
    sys.exit(1)  # fire atexit hooks


signal.signal(signal.SIGTERM, terminate_child)


def unique_dir(dstdir=None, hint=''):  # defaults to /tmp
    hint = hint if len(hint) <= 20 else hint[:20-2] + '..'
    return tempfile.mkdtemp(prefix=('.' + hint + '.'), dir=dstdir)


def remove(*paths):
    for path in paths:
        assert AUTOREMOVE_PREFIX in path
        try:
            if not os.path.isdir(path) or os.path.islink(path):
                try:
                    os.unlink(path)
                except FileNotFoundError:
                    pass
            else:
                shutil.rmtree(path, ignore_errors=True)
        except Exception as e:
            log.warning(f'cleanup error for {path}: {e}')


def disappearing_file(dstdir=None, hint=''):
    prefix = AUTOREMOVE_PREFIX + hint + '.' if hint else AUTOREMOVE_PREFIX
    _, temp_file_path = tempfile.mkstemp(prefix=prefix, dir=dstdir)
    assert AUTOREMOVE_PREFIX in temp_file_path
    atexit.register(lambda: remove(temp_file_path))
    return temp_file_path


def disappearing_dir(dstdir=None, hint=''):
    prefix = AUTOREMOVE_PREFIX + hint + '.' if hint else AUTOREMOVE_PREFIX
    temp_dir_path = tempfile.mkdtemp(prefix=prefix, dir=dstdir)
    assert AUTOREMOVE_PREFIX in temp_dir_path
    atexit.register(lambda: remove(temp_dir_path))
    return temp_dir_path


def has_space(how_much='2G', reserve_fraction=.3, where='/tmp'):
    for suffix, power in {'G': 30, 'M': 20, 'K': 10}.items():
        if isinstance(how_much, str) and how_much.endswith(suffix):
            how_much = float(how_much[:-1]) * 2 ** power
            break
    total, _, free = shutil.disk_usage(where)
    if not free >= how_much:
        log.warning(f'{where} does not have {how_much} of free space')
    if not free >= total * reserve_fraction:
        log.warning(f'{where} is {int((1 - reserve_fraction) * 100)}% full')
    return free >= how_much and free >= total * reserve_fraction



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import itertools
import os
import sys

import fingertip
import fingertip.util.cleanup_job
import fingertip.util.log
import fingertip.util.path
import fingertip.util.reflink

from fingertip.util import optional_pretty_backtraces  # noqa: F401


def parse_kwarg(kwarg):
    key, val = kwarg.split('=', 1) if '=' in kwarg else (kwarg, True)
    return key.replace('-', '_'), val


def parse_subcmd(subcmd, *all_args):
    args = [a for a in all_args if not a.startswith('--')]
    kwargs = [parse_kwarg(a[2:])
              for a in sorted(all_args) if a.startswith('--')]
    return subcmd, args, dict(kwargs)


def main():
    # Start with plain to get output from setup wizard
    fingertip.util.log.plain()

    fingertip.util.reflink.storage_setup_wizard()
    fingertip.util.cleanup_job.schedule()

    fingertip.util.log.nicer()

    args = sys.argv[1:]

    if not args:
        fingertip.util.log.error('no plugin specified')
        sys.exit(1)

    subcmds = [list(ws)
               for x, ws in itertools.groupby(args, lambda w: w != '+') if x]
    first_step, *rest_of_the_steps = [parse_subcmd(*sc) for sc in subcmds]

    first_step_cmd, first_step_args, first_step_kwargs = first_step
    m = fingertip.build(first_step_cmd, *first_step_args, **first_step_kwargs,
                        fingertip_last_step=(not rest_of_the_steps))

    for i, (step_cmd, step_args, step_kwargs) in enumerate(rest_of_the_steps):
        last_step = i == len(rest_of_the_steps) - 1
        m = m.apply(step_cmd, *step_args, **step_kwargs,
                    fingertip_last_step=last_step)

    if m:
        success_log = m
        fingertip.util.log.plain()
        DEBUG = os.getenv('FINGERTIP_DEBUG') == '1'
        msg = (f'For more details, check {success_log} '
               'or set FINGERTIP_DEBUG=1.'
               if not DEBUG else f'Logfile: {success_log}')
        fingertip.util.log.info(f'Success. {msg}')
    else:
        msg = f'Success, no log. Rerun with FINGERTIP_DEBUG=1 for more info.'


if __name__ == '__main__':
    main()



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import fingertip
import fingertip.util.temp


@fingertip.transient
def main(m):
    scriptname = fingertip.util.temp.disappearing_file()
    m.log.debug(f'Scriptname: {scriptname}')
    filename = '/etc/TADAA'
    with open(scriptname, "w") as f:
        f.write(f'#!/bin/sh\n\ntouch {filename}')
    with m.apply('script.run', scriptname) as m:
        assert m(f'ls -1 {filename}').out.strip() == filename



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import sys
import fingertip


# script.X (X in [run, setup]) without --cache=Y defaults to:
#  * not caching if it is the last plugin in the pipeline
#  * caching for just one invocation
#    if there are further plugins in the pipeline
# Examples:
# fingertip ... + script.X script  -  no cache, never persist, always rerun
# fingertip ... + script.X script + ssh  -  cache just for this invocation
# fingertip ... + script.X script --cache=1h  -  cache for 1h at most
# fingertip ... + script.X script --cache=1h + ssh  -  cache or reuse if fresh
# fingertip ... + transient script.X script + ssh  -  revert
def _should_be_transient(m, scriptpath, cache=0, no_unseal=False):
    return False if cache else 'last'


def _exec(m, scriptpath, mode=None, last_step=None, cache=0, no_unseal=False):
    m.log.debug(f'Plugin: script, mode: {mode}, scriptpath: {scriptpath}, '
                f'last_step: {last_step}, cache: {cache}, '
                f'no_unseal: {no_unseal}')
    assert mode in ['run', 'test']
    m = m if no_unseal else m.apply('unseal')
    scriptname = 'uploaded_script'

    with m:
        if cache is not True:
            m.expiration.cap(cache)
        m.expiration.depend_on_a_file(scriptpath)

        m.log.debug(f'uploading {scriptpath} into {scriptname}')
        m.ssh.upload(scriptpath, dst=scriptname)
        m.log.debug(f'uploaded {scriptpath} into {scriptname}')
        m(f'chmod +x {scriptname}')
        if mode == 'test':
            m.hooks.disable_proxy()
            check = False
        elif mode == 'run':
            check = True
        else:
            assert False
        m.log.plain()
        exit_code = m(f'./{scriptname}', check=check).retcode
        m.log.nicer()
        if mode == 'test' and exit_code != 0:
            m.log.warning(f'failed with exit code: {exit_code}')
            if last_step:
                sys.exit(exit_code)
    return m

# execute setup-like scenario, meaning mostly
#  * do not touch proxy settings
#  * traceback on non-zero exit code
@fingertip.transient(when=_should_be_transient)
def run(m, scriptpath, fingertip_last_step=None, cache=0, no_unseal=False):
    return _exec(m, scriptpath, mode='run',
                 last_step=fingertip_last_step,
                 cache=cache, no_unseal=no_unseal)

# run test like scenario, meaning mostly:
#  * disable proxy
#  * do not traceback on non-zero exit code
@fingertip.transient(when=_should_be_transient)
def test(m, scriptpath, fingertip_last_step=None, cache=0, no_unseal=False):
    return _exec(m, scriptpath, mode='test',
                 last_step=fingertip_last_step,
                 cache=cache, no_unseal=no_unseal)



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import datetime
import numbers
import os
import sys
import time

from fingertip.util import log, weak_hash


def _parse(interval):
    _SUFFIXES = {'s': 1, 'm': 60, 'h': 60 * 60, 'd': 24 * 60 * 60}
    if isinstance(interval, str) and interval[-1] in _SUFFIXES:
        return float(interval[:-1]) * _SUFFIXES[interval[-1]]
    if isinstance(interval, numbers.Real) and not isinstance(interval, bool):
        return float(interval)
    raise ValueError(f'Cannot parse time interval {interval}')


class Expiration:
    def __init__(self, expire_in):
        self.time = time.time() + _parse(expire_in)
        self._deps = {}

    def pretty(self):
        return datetime.datetime.fromtimestamp(self.time).isoformat()

    def cap(self, interval):
        self.time = min(self.time, time.time() + _parse(interval))

    def is_expired(self, by=None):
        return self.time < (by or time.time())

    def depend_on_a_file(self, path):
        path = os.path.abspath(path)
        if path.startswith('/usr/lib') or '/site-packages/' in path:
            return
        self._deps[path] = (os.stat(path).st_mtime, weak_hash.of_file(path))

    def files_have_not_changed(self):
        if os.getenv('FINGERTIP_IGNORE_CODE_CHANGES', '0') != '0':
            return True
        for path, (mtime, hash_) in self._deps.items():
            log.debug(f'checking that {path} has not changed...')
            if mtime != (os.stat(path).st_mtime):
                if hash_ != weak_hash.of_file(path):
                    log.warning(f'{path} has changed, set '
                                'FINGERTIP_IGNORE_CODE_CHANGES=1 to ignore')
                    return False
        return True

    def depend_on_loaded_python_modules(self):
        """
        Make the machine depend on all python modules loaded to date
        in an imprecise best-effort overly-cautious attempt
        to get automatic rebuilds triggered on source code changes.
        """
        for module in sys.modules:
            try:
                module_file = sys.modules[module].__file__
                if module_file:
                    self.depend_on_a_file(module_file)
            except AttributeError:
                pass



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.


try:
    import backtrace
    styles = backtrace.STYLES.copy()
    styles['call'] = styles['call'].replace('--> ', '')
    backtrace.hook(strip_path=True, on_tty=True, styles=styles)
except ImportError:
    pass



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import fasteners
import git

import os
import tarfile


from fingertip.util import log, path, temp


OFFLINE = os.getenv('FINGERTIP_OFFLINE', '0') != '0'
DIR = path.downloads('git')


class Repo(git.Repo):
    def __init__(self, url, *path_components, enough_to_have=None):
        self.url = url
        self.path = os.path.join(DIR, *path_components)
        lock_path = self.path + '-lock'
        self.lock = fasteners.process_lock.InterProcessLock(lock_path)
        self.lock.acquire()
        if not os.path.exists(self.path):
            log.info(f'cloning {url}...')
            git.Repo.clone_from(url, self.path, mirror=True)  # TODO: use bare
            super().__init__(self.path)
        else:
            super().__init__(self.path)
            update_not_needed = enough_to_have and (
                enough_to_have in (t.name for t in self.tags) or
                enough_to_have in (h.name for h in self.heads) or
                enough_to_have in (c.hexsha for c in self.iter_commits())
                # that's not all commits, but best-effort should be fine here
            )
            if update_not_needed:
                log.info(f'not re-fetching {url} '
                         f'because {enough_to_have} is already present')
            if OFFLINE:
                log.info(f'not re-fetching {url} because of offline mode')
            if not OFFLINE and not update_not_needed:
                log.info(f'updating {url}...')
                self.remote().fetch(tags=True)
        self.lock.release()

    def __enter__(self):
        self.lock.acquire()
        return self

    def __exit__(self, *_):
        self.lock.release()


def cached_clone(m, url, path_in_m, rev=None, rev_is_enough=True):
    assert hasattr(m, 'ssh')
    with m:
        kwa = {} if not rev_is_enough else {'enough_to_have': rev}
        with Repo(url, url.replace('/', '::'), **kwa) as repo:
            tar = temp.disappearing_file()
            tar_in_m = f'/tmp/{os.path.basename(tar)}'
            extracted_in_m = f'/tmp/{os.path.basename(tar)}-extracted'
            log.info(f'packing {url} checkout...')
            with tarfile.open(tar, 'w') as tf:
                tf.add(repo.path, arcname=extracted_in_m)
            log.info(f'uploading {url} checkout...')
            m.ssh.upload(tar, tar_in_m)
        log.info(f'performing {url} checkout...')
        m(f'''
            set -uex
            tar xmvf {tar_in_m} -C /
            mkdir -p {path_in_m}
            git clone -n {extracted_in_m} {path_in_m}
            cd {path_in_m}
            git remote set-url origin {url}
            git checkout {f'{rev}' if rev else 'origin/HEAD'}
            rm -rf {extracted_in_m}
            rm -f {tar_in_m}
        ''')
    return m



"""
Helper functions for fingertep: weak hashing.
"""

import hashlib


def of_string(s):
    return hashlib.sha256(s.encode()).hexdigest()[:8]


def of_file(fname):
    h = hashlib.sha256()
    with open(fname, 'rb') as f:
        while True:
            d = f.read(65563)
            if not d:
                return h.hexdigest()[:8]
            h.update(d)



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import os

import fingertip
from fingertip.util import path


MIRROR = 'http://dl-cdn.alpinelinux.org/alpine'
ISO = MIRROR + '/v3.10/releases/x86_64/alpine-virt-3.10.3-x86_64.iso'
REPO = MIRROR + '/v3.10/main/'


def main(m=None):
    m = m or fingertip.build('backend.qemu', ram_size='128M')
    if hasattr(m, 'qemu'):
        m = m.apply(install_in_qemu).apply(first_boot)
    elif hasattr(m, 'container'):
        # podman-criu: https://github.com/checkpoint-restore/criu/issues/596
        m = m.apply(m.container.from_image, 'alpine')
    else:
        raise NotImplementedError()

    with m:
        def prepare():
            m('apk add python3')
        m.hooks.ansible_prepare.append(prepare)
        #m.hooks.ansible_prepare.append(lambda: m('apk add python3'))
    return m


def install_in_qemu(m):
    iso_file = os.path.join(m.path, os.path.basename(ISO))
    m.http_cache.fetch(ISO, iso_file)

    with m:
        m.qemu.run(load=None, extra_args=['-cdrom', iso_file])
        m.console.expect_exact('localhost login: ')
        m.console.sendline('root')
        m.console.expect_exact('localhost:~# ')

        m.hostname = 'alpine'
        m.prompt = f'{m.hostname}:~# '

        m.console.sendline('setup-alpine -q')
        m.console.expect_exact('Select keyboard layout [none]:')
        m.console.sendline('us')
        m.console.expect_exact('Select variant []:')
        m.console.sendline('us')
        m.console.expect_exact(m.prompt)

        m.console.sendline(f'setup-proxy {m.http_cache.internal_url}')
        m.console.expect_exact(m.prompt)
        m.console.sendline('. /etc/profile.d/proxy.sh')
        m.console.expect_exact(m.prompt)

        fqdn = m.hostname + '.fingertip.local'
        m.console.sendline(f'echo "{m.hostname}" > /etc/hostname')
        m.console.expect_exact(m.prompt)
        m.console.sendline('hostname -F /etc/hostname')
        m.console.expect_exact(m.prompt)
        m.console.sendline(f'echo "127.0.0.1 {fqdn} {m.hostname}" > /etc/hosts')
        m.console.expect_exact(m.prompt)

        m.console.sendline(f'setup-apkrepos {REPO}')
        m.console.expect_exact(m.prompt)

        m.console.sendline('setup-sshd -c openssh')
        m.console.expect_exact(m.prompt)

        m.console.sendline('setup-disk -m sys /dev/vda')
        m.console.expect_exact('Erase the above disk(s) and continue? [y/N]:')
        m.console.sendline('y')
        m.console.expect_exact(m.prompt)

        m.console.sendline('fstrim -v /')
        m.console.expect_exact(m.prompt)

        m.console.sendline('poweroff')

        m.qemu.wait()
        m.qemu.compress_image()

        def disable_proxy():
            m('setup-proxy none', check=False)
            m.console.sendline(f'unset http_proxy https_proxy ftp_proxy')
            m.console.expect_exact(m.prompt)
        m.hooks.disable_cache.append(disable_proxy)

        return m


def first_boot(m):
    ssh_key_fname = path.fingertip('ssh_key', 'fingertip.pub')
    with open(ssh_key_fname) as f:
        ssh_pubkey = f.read().strip()

    with m:
        m.qemu.run(load=None)
        m.console.expect_exact(f'{m.hostname} login: ')
        m.console.sendline('root')
        m.console.expect_exact(m.prompt)

        m.console.sendline(f'install -m 700 -d .ssh')
        m.console.expect_exact(m.prompt)
        m.console.sendline(f'echo "{ssh_pubkey}" >> .ssh/authorized_keys')
        m.console.expect_exact(m.prompt)
        m.expiration.depend_on_a_file(ssh_key_fname)

        m.hooks.unseal.append(lambda: m('/etc/init.d/networking restart'))
    return m



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

# setsebool -P container_manage_cgroup true

import logging
import os
import subprocess
import sys
import time

import pexpect

# TODO: sealing
# TODO: http proxy for podman
# TODO: http proxy for container itself
import fingertip.exec
import fingertip.machine
import fingertip.util.http_cache
from fingertip.util import reflink, repeatedly, weak_hash


def _podman(log, *args, func='run', **kwargs):
    if func not in ('check_output', 'Popen'):
        kwargs['check'] = True
    redirects = {}
    if func != 'check_output' and 'stdout' not in kwargs:
        redirects['stdout'] = logging.INFO
    if 'stderr' not in kwargs:
        redirects['stderr'] = logging.ERROR
    func = getattr(subprocess, func)
    func = log.pipe_powered(func, **redirects)
    return func(('sudo', 'podman',) + args, **kwargs)


def _copy_ownership(log, p1, p2):
    uid, gid = os.stat(p1).st_uid, os.stat(p1).st_gid
    run = log.pipe_powered(subprocess.run,
                           stdout=logging.INFO, stderr=logging.ERROR)
    run(['sudo', 'chown', f'{uid}:{gid}', p2], check=True)


def _base():
    # TODO: seal
    m = fingertip.machine.Machine('podman-criu', sealed=False, expire_in='4h')
    m.container = ContainerNamespacedFeatures(m)
    m.container.from_image = from_image
    m._backend_mode = 'pexpect'

    def up():
        m.log.debug('up')
        m.container.name = 'fingertip_' + weak_hash.of_string(m.path)
        if not hasattr(m.container, 'starting_image'):  # no starting image yet
            return
        # need to load the initial image, checkpoint and create image.tar
        m.container.container_id = _podman(
                m.log,
                'container', 'restore', '-n', m.container.name,
                '-i', os.path.join(m.path, 'snapshot.tar'),
                func='check_output'
        ).decode().strip()
        assert m.container.container_id
        m.log.debug(f'restore -> container_id = {m.container.container_id}')
        if m._backend_mode == 'pexpect':
            pexp = m.log.pseudofile_powered(pexpect.spawn,
                                            logfile=logging.INFO)
            m.console = pexp('sudo', ['podman', 'attach',
                                      m.container.container_id],
                             echo=False, timeout=None,
                             encoding='utf-8', codec_errors='ignore')
        elif m._backend_mode == 'direct':
            subprocess.run(['sudo', 'podman', 'attach',
                           m.container.container_id])
    m.hooks.up.append(up)

    def detach(retries=10, timeout=1/32):
        def dtch():
            time.sleep(.1)  # HACK, FIXME
            m.console.sendcontrol('p')
            time.sleep(.001)  # HACK
            m.console.sendcontrol('q')
            m.console.expect(pexpect.EOF, timeout=.1)
        repeatedly.keep_trying(dtch, pexpect.exceptions.TIMEOUT,
                               retries=retries, timeout=timeout)
        m.console.wait()
    m.hooks.detach.append(detach)

    def down():
        if not hasattr(m.container, 'starting_image'):  # no starting image yet
            return
        if m.console:
            m.hooks.detach()
        m.console = None
        _podman(m.log, 'container', 'checkpoint', m.container.container_id,
                '-e', os.path.join(m.path, 'snapshot.tar'))
        _copy_ownership(m.log, m.path, os.path.join(m.path, 'snapshot.tar'))
        _podman(m.log, 'rm', m.container.container_id)
        del m.container.container_id
        del m.container.name
    m.hooks.down.append(down)

    def drop():
        if not hasattr(m.container, 'starting_image'):  # no starting image yet
            return
        if m.console:
            m.hooks.detach()
        m.console = None
        _podman(m.log, 'stop', '-t', '0', m.container.container_id)
        _podman(m.log, 'rm', m.container.container_id)
        del m.container.container_id
        del m.container.name
    m.hooks.drop.append(drop)

    def clone(to):
        if not hasattr(m.container, 'starting_image'):  # no starting image yet
            return
        m.log.debug(f'{m} {to}')
        reflink.auto(os.path.join(m.path, 'snapshot.tar'),
                     os.path.join(to, 'snapshot.tar'))
    m.hooks.clone.append(clone)

    def attach_exec():
        m.exec = m.container.exec
    m.hooks.up.append(attach_exec)  # FIXME self_test.exec, I blame cloudpickle

    return m


def main(image=None):
    m = fingertip.build(_base)
    if image:
        m = m.apply(from_image, image)
    return m


class ContainerNamespacedFeatures:
    def __init__(self, m):
        self._args = []
        self.m = m

    def exec(self, *cmd, shell=False):
        cmd = ('sh', '-c', *cmd) if shell else cmd
        stdout = self.m.log.make_pipe(level=logging.INFO)
        stderr = self.m.log.make_pipe(level=logging.INFO)
        try:
            p = _podman(self.m.log, 'exec', self.container_id, *cmd,
                        stdout=stdout, stderr=stderr, func='Popen')
            p.wait()
        finally:
            stdout.close()
            stderr.close()
            stdout.wait()
            stderr.wait()
        out, err = stdout.data, stderr.data
        if p.returncode:  # HACK: ugly workaround for podman polluting stderr
            m = (f'Error: non zero exit code: {p.returncode}:'
                 ' OCI runtime error\n').encode()
            if err.endswith(m):
                err = err[:-len(m)]
        return fingertip.exec.ExecResult(p.returncode, out, err)


def from_image(m=None, image=None, cmd=[]):
    m = m or fingertip.build(main)
    with m:
        m.container.name = 'fingertip_' + weak_hash.of_string(m.path)
        m.container.starting_image = image
        m.container.container_id = _podman(
                m.log,
                'run', '-dit', '--name', m.container.name, *m.container._args,
                image, *cmd,
                func='check_output'
        ).decode().strip()
        m.log.debug(f'run -> container_id = {m.container.container_id}')
        _podman(m.log, 'container', 'checkpoint', m.container.container_id,
                '-e', os.path.join(m.path, 'snapshot.tar'))
        _copy_ownership(m.log, m.path, os.path.join(m.path, 'snapshot.tar'))
        _podman(m.log, 'container', 'rm', m.container.container_id)
        m.hooks.up()
        m.console.sendline(' echo TEST')
        m.console.expect_exact('echo TEST\r\n')
        m.console.expect_exact('TEST\r\n')
        m.console.sendline('')
        m.console.expect_exact('\r\n')
        m.console.sendline('')
        m.console.expect_exact('\r\n')
        m.prompt = m.console.before
        m.log.debug(f'm.prompt: "{m.prompt}"')
        m.console.sendline('')
        m.console.expect_exact(m.prompt)
        return m



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import requests

import os

import fingertip.machine
from fingertip.util import path


FEDORA_GEOREDIRECTOR = 'http://download.fedoraproject.org/pub/fedora/linux'


def determine_mirror(mirror):
    h = requests.head(mirror, allow_redirects=False)
    if h.status_code in (301, 302, 303, 307, 308) and 'Location' in h.headers:
        return h.headers['Location'].rstrip('/')
    return mirror


def main(m=None, version=31, updates=True,
         mirror=None, resolve_redirect=False):
    m = m or fingertip.build('backend.qemu')
    if hasattr(m, 'qemu'):
        m = m.apply(install_in_qemu, version=version, updates=updates,
                    mirror=mirror, resolve_redirect=resolve_redirect)
    elif hasattr(m, 'container'):
        m = m.apply(m.container.from_image, f'fedora:{version}')
        if updates:
            with m:
                m('dnf -y update')
    else:
        raise NotImplementedError()
    return m


def install_in_qemu(m, version, updates=True,
                    mirror=None, resolve_redirect=False):
    ml_norm = ('http://mirrors.fedoraproject.org/metalink' +
               f'?repo=fedora-{version}&arch=x86_64&protocol=http')
    ml_upd = ('http://mirrors.fedoraproject.org/metalink' +
              f'?repo=updates-released-f{version}&arch=x86_64&protocol=http')
    releases_development = 'development' if version == '32' else 'releases'
    if mirror:
        if resolve_redirect:
            m.log.info(f'autoselecting mirror by redirect from {mirror}...')
            mirror = determine_mirror(mirror)
        url = f'{mirror}/{releases_development}/{version}/Everything/x86_64/os'
        upd = f'{mirror}/updates/{version}/Everything/x86_64'
        repos = (f'url --url {url}\n' +
                 f'repo --name fedora --baseurl {url}\n' +
                 (f'repo --name updates --baseurl {upd}'
                  if updates else ''))
    else:
        m.log.info('autoselecting mirror...')
        mirror = determine_mirror(FEDORA_GEOREDIRECTOR)
        url = f'{mirror}/{releases_development}/{version}/Everything/x86_64/os'
        repos = (f'url --url {url}\n' +
                 f'repo --name fedora --metalink {ml_norm}\n' +
                 (f'repo --name updates --metalink {ml_upd}'
                  if updates else ''))
    m.log.info(f'selected mirror: {mirror}')

    m.expiration.cap('2d')  # non-immutable repositories

    original_ram_size = m.qemu.ram_size

    with m:
        m.qemu.ram_size = '2G'
        hostname = f'fedora{version}'
        fqdn = hostname + '.fingertip.local'

        ssh_key_fname = path.fingertip('ssh_key', 'fingertip.pub')
        with open(ssh_key_fname) as f:
            ssh_pubkey = f.read().strip()
        m.expiration.depend_on_a_file(ssh_key_fname)

        ks_fname = path.fingertip('kickstart_templates', f'fedora{version}')
        with open(ks_fname) as f:
            ks_text = f.read().format(HOSTNAME=fqdn,
                                      SSH_PUBKEY=ssh_pubkey,
                                      PROXY=m.http_cache.internal_url,
                                      REPOS=repos)
        m.expiration.depend_on_a_file(ks_fname)

        m.http_cache.mock('http://ks', text=ks_text)
        m.log.info(f'fetching kernel: {url}/isolinux/vmlinuz')
        kernel = os.path.join(m.path, 'kernel')
        m.http_cache.fetch(f'{url}/isolinux/vmlinuz', kernel)
        m.log.info(f'fetching initrd: {url}/isolinux/initrd.img')
        initrd = os.path.join(m.path, 'initrd')
        m.http_cache.fetch(f'{url}/isolinux/initrd.img', initrd)
        append = ('ks=http://ks inst.ksstrict console=ttyS0 inst.notmux '
                  f'proxy={m.http_cache.internal_url} ' +
                  f'inst.proxy={m.http_cache.internal_url} ' +
                  f'inst.repo={url}')
        extra_args = ['-kernel', kernel, '-initrd', initrd, '-append', append]

        m.qemu.run(load=None, extra_args=extra_args)
        i = m.console.expect(['Storing configuration files and kickstarts',
                              'installation failed',
                              'installation was stopped',
                              'installer will now terminate'])
        assert i == 0, 'Installation failed'
        m.qemu.wait()
        m.qemu.compress_image()
        m.qemu.ram_size = original_ram_size
        m.qemu.run(load=None)  # cold boot
        ROOT_PASSWORD = 'fingertip'
        m.prompt = f'[root@{hostname} ~]# '
        m.console.expect(f'{hostname} login: ')
        m.console.sendline('root')
        m.console.expect('Password: ')
        m.console.sendline(ROOT_PASSWORD)
        m.console.expect_exact(m.prompt)
        m.log.info('Fedora installation finished')

        def disable_proxy():
            return m.apply('ansible', 'ini_file', path='/etc/dnf/dnf.conf',
                           section='main', option='proxy', state='absent')
        m.hooks.disable_proxy.append(disable_proxy)

        m.hooks.unseal += [lambda: m('systemctl restart NetworkManager'),
                           lambda: m('nm-online')]

        m.fedora = version

        return m



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

"""
A plugin to execute Ansible playbooks and ad-hoc commands.

Ansible's Python API is not stable yet, so... CLI.
"""

import logging
import os
import subprocess

from fingertip.util import temp


def prepare(m):
    if not hasattr(m, '_ansible_prepare') and m.hooks.ansible_prepare:
        with m:
            m.hooks.ansible_prepare()
            m._ansible_prepare = True
    return m


def _ansible(m, *args, check=True, cmd=('ansible', 'fingertip')):
    env = {**os.environ, 'ANSIBLE_HOST_KEY_CHECKING': 'False'}
    if hasattr(m, 'ssh'):
        m.ssh.exec('true')  # to ensure correct spin-up
        connection = 'ssh'  # TODO: compare with paramiko
        prefix = ()
        host = ['fingertip',
                'ansible_connection=ssh',
                'ansible_user=root',
                'ansible_host=localhost',
                f'ansible_port={m.ssh.port}',
                f'ansible_ssh_private_key_file={m.ssh.key_file}']
    elif m.backend == 'podman-criu':
        connection = 'podman'
        prefix = ('sudo', '-H')
        host = ['fingertip', f'ansible_host={m.container.container_id}']
    else:
        raise NotImplementedError()
    inventory = temp.disappearing_file(hint='ansible-inventory')
    with open(inventory, 'w') as f:
        f.write(' '.join(host))
    more_opts = ('-T', '120', '-i', inventory, '-c', connection)
    cmd = prefix + cmd + more_opts + args
    m.log.debug(' '.join(cmd))
    run = m.log.pipe_powered(subprocess.run,
                             stdout=logging.INFO, stderr=logging.INFO)
    return run(cmd, env=env, check=check)


def main(m, module, *args, **kwargs):
    def to_str(v):
        if isinstance(v, list) or isinstance(v, tuple):
            return ','.join(v)
        if isinstance(v, bool):
            return 'yes' if v else 'no'
        return str(v)
    module_args = args + tuple(f'{k}="{to_str(v)}"' for k, v in kwargs.items())
    with m.apply(prepare) as m:
        _ansible(m, '-m', module, '-a', ' '.join(module_args))
    return m


def playbook(m, playbook_path):
    with m.apply(prepare) as m:
        _ansible(m, playbook_path, cmd=('ansible-playbook',))
        # not comprehensive; best-effort
        m.expiration.depend_on_a_file(playbook_path)
    return m



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import hashlib
import http.server
import os
import shutil
import socketserver
import threading

import requests
import requests_mock
import cachecontrol
import cachecontrol.caches

from fingertip.util import path, log


OFFLINE = os.getenv('FINGERTIP_OFFLINE', '0') != '0'
BIG = 2**30  # too big for caching
STRIP_IF_OFFLINE = ('Cache-Control', 'Pragma')
STRIP_ALWAYS = ('TE', 'Transfer-Encoding', 'Keep-Alive', 'Trailer', 'Upgrade',
                'Connection', 'Host', 'Accept')
STRIP_HEADERS = STRIP_ALWAYS + STRIP_IF_OFFLINE if OFFLINE else STRIP_ALWAYS


class ThreadingHTTPServer(socketserver.ThreadingMixIn, http.server.HTTPServer):
    pass


class HTTPCache:
    def __init__(self, host='127.0.0.1', port=0):
        self.host = host
        self._mocks = []
        http_cache = self

        class Handler(http.server.SimpleHTTPRequestHandler):
            protocol_version = 'HTTP/1.1'

            def _status_and_headers(self, status_code, headers):
                self.send_response(status_code)
                for k, v in headers.items():
                    self.send_header(k, v)
                self.end_headers()

            def _serve(self, uri, headers, meth='GET'):
                sess = http_cache._get_requests_session()

                headers = {k: v for k, v in headers.items() if
                           not (k in STRIP_HEADERS or k.startswith('Proxy-'))}
                log.debug(f'{meth} {uri}')
                for k, v in headers.items():
                    log.debug(f'{k}: {v}')

                try:
                    if meth == 'GET' and not OFFLINE:
                        # direct streaming might be required...
                        preview = sess.head(uri, headers=headers,
                                            allow_redirects=False)
                        direct = None
                        if int(preview.headers.get('Content-Length', 0)) > BIG:
                            direct = f'file bigger than {BIG}'
                        if 'Range' in headers:
                            # There seems to be a bug in CacheControl
                            # that serves contents in full if a range request
                            # hits a non-ranged cached entry.
                            direct = f'ranged request, playing safe'
                        if direct:
                            # Don't cache, don't reencode, stream it as is
                            log.warning(f'streaming {uri} directly ({direct})')
                            r = requests.get(uri, headers=headers, stream=True)
                            self._status_and_headers(r.status_code, r.headers)
                            self.copyfile(r.raw, self.wfile)
                            return

                    # fetch with caching
                    m_func = getattr(sess, meth.lower())
                    r = m_func(uri if '://' in uri else 'http://self' + uri,
                               headers=headers, allow_redirects=False)
                    data = r.content
                    length = int(r.headers.get('Content-Length', 0))
                    if len(data) != length:
                        data = hack_around_unpacking(uri, headers, data)
                    assert len(data) == length
                    self._status_and_headers(r.status_code, r.headers)
                    if meth == 'GET':
                        self.wfile.write(data)
                    log.info(f'{meth} {uri} served {length}')
                except BrokenPipeError:
                    log.warning(f'Broken pipe for {meth} {uri}')
                except ConnectionResetError:
                    log.warning(f'Connection reset for {meth} {uri}')
                except requests.exceptions.ConnectionError:
                    log.warning(f'Connection error for {meth} {uri}')

            def do_HEAD(self):
                self._serve(uri=self.path, headers=self.headers, meth='HEAD')

            def do_GET(self):
                self._serve(uri=self.path, headers=self.headers, meth='GET')

            def log_message(self, format, *args):  # supress existing logging
                return

        httpd = ThreadingHTTPServer((host, port), Handler)
        _, self.port = httpd.socket.getsockname()
        threading.Thread(target=httpd.serve_forever, daemon=True).start()

    def _get_requests_session(self, direct=False):
        if not direct:
            cache = cachecontrol.caches.FileCache(path.downloads('cache'))
            sess = cachecontrol.CacheControl(requests.Session(), cache=cache)
        else:
            sess = requests.Session()
        for uri, kwargs in self._mocks:
            adapter = requests_mock.Adapter()
            adapter.register_uri('HEAD', uri, **kwargs)
            adapter.register_uri('GET', uri, **kwargs)
            sess.mount(uri, adapter)
        return sess

    def fetch(self, url, out_path):
        sess = self._get_requests_session()
        r = sess.get(url)
        with open(out_path, 'wb') as f:
            f.write(r.content)

    def mock(self, uri, text):
        """
        Examples:
          * mock('http://self/test', 'TEST')
            and access directly as `http://<host>:<port>/test`
          * mock('http://anything/test', 'ANYTHINGT')
            and access through proxy as `http://anything/test`
        """
        content_length = {'Content-Length': str(len(text.encode()))}
        self._mocks.append((uri, {'text': text, 'headers': content_length}))


# Hack around requests uncompressing Content-Encoding: gzip


def hack_around_unpacking(uri, headers, wrong_content):
    log.warning(f're-fetching correct content for {uri}')
    r = requests.get(uri, headers=headers, stream=True, allow_redirects=False)
    h = hashlib.sha256(wrong_content).hexdigest()
    cachefile = path.downloads('fixups', h, makedirs=True)
    if not os.path.exists(cachefile):
        with path.wip(cachefile) as wip:
            with open(wip, 'wb') as f:
                shutil.copyfileobj(r.raw, f)
    with open(cachefile, 'rb') as f:
        return f.read()


# Fully offline cache utilization functionality


def c_r_offline(self, request):
    cache_url = self.cache_url(request.url)
    log.debug(f'looking up {cache_url} in the cache')
    cache_data = self.cache.get(cache_url)
    if cache_data is None:
        log.error(f'{cache_url} not in cache and fingertip is offline')
        return False
    resp = self.serializer.loads(request, cache_data)
    if not resp:
        log.error(f'{cache_url} cache entry deserialization failed, ignored')
        return False
    log.warning(f'Using {cache_url} from offline cache')
    return resp


if OFFLINE:
    cachecontrol.controller.CacheController.cached_request = c_r_offline



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import json
import logging
import os
import selectors
import socket
import stat
import subprocess
import time

import fasteners
import pexpect

import fingertip.machine
import fingertip.util.http_cache
from fingertip.util import free_port, log, path, reflink, repeatedly, temp


CACHE_INTERNAL_IP, CACHE_INTERNAL_PORT = '10.0.2.244', 8080
CACHE_INTERNAL_URL = f'http://{CACHE_INTERNAL_IP}:{CACHE_INTERNAL_PORT}'
# TODO: add a way to customize smp
QEMU_COMMON_ARGS = ['-enable-kvm', '-cpu', 'host,-vmx', '-smp', '4',
                    '-virtfs', f'local,id=shared9p,path={path.SHARED},'
                               'security_model=mapped-file,mount_tag=shared',
                    '-nographic',
                    '-object', 'rng-random,id=rng0,filename=/dev/urandom',
                    '-device', 'virtio-rng-pci,rng=rng0']


def main(arch='x86_64', ram_size='1G', disk_size='20G',
         custom_args=[], guest_forwards=[]):
    assert arch == 'x86_64'
    # FIXME: -tmp
    m = fingertip.machine.Machine('qemu')
    m.arch = arch
    m.qemu = QEMUNamespacedFeatures(m, ram_size, disk_size, custom_args)
    m._backend_mode = 'pexpect'

    def load():
        m.http_cache = fingertip.util.http_cache.HTTPCache()
        m.http_cache.internal_ip = CACHE_INTERNAL_IP
        m.http_cache.internal_port = CACHE_INTERNAL_PORT
        m.http_cache.internal_url = CACHE_INTERNAL_URL
    m.hooks.load.append(load)

    def up():
        if m.qemu.live:
            m.qemu.run()
    m.hooks.up.append(up)

    def down():
        if m.qemu.live:
            m.log.debug(f'save_live {m.qemu.monitor}, {m.qemu.monitor._sock}')
            m.qemu.monitor.pause()
            m.qemu.monitor.checkpoint()
            m.qemu.monitor.commit()
            m.qemu.monitor.quit()
            m.qemu._go_down()
    m.hooks.down.append(down)

    def drop():
        if m.qemu.live:
            m.log.debug(f'drop {m.qemu.monitor}, {m.qemu.monitor._sock}')
            m.qemu.monitor.quit()
            m.qemu._go_down()
    m.hooks.drop.append(drop)

    def save():
        m.http_cache = None
    m.hooks.save.append(save)

    def clone(to_path):
        m.qemu._image_to_clone = os.path.join(m.path, 'image.qcow2')
    m.hooks.clone.append(clone)

    def disrupt():
        if m.qemu.live:
            m.ssh.invalidate()
    m.hooks.disrupt.append(disrupt)

    m.breakpoint = lambda: m.apply('ssh', no_unseal=True)

    load()
    run = m.log.pipe_powered(subprocess.run,
                             stdout=logging.INFO, stderr=logging.ERROR)
    run(['qemu-img', 'create', '-f', 'qcow2',
         os.path.join(m.path, 'image.qcow2'), disk_size], check=True)
    return m


class QEMUNamespacedFeatures:
    def __init__(self, vm, ram_size, disk_size, custom_args):
        self.vm = vm
        self.live = False
        self.ram_size, self.disk_size = ram_size, disk_size
        self.custom_args = custom_args
        self._image_to_clone = None
        self._qemu = f'qemu-system-{self.vm.arch}'

    def run(self, load='tip', guest_forwards=[], extra_args=[]):
        run_args = ['-loadvm', load] if load else []

        self.monitor = Monitor(self.vm)
        run_args += ['-qmp', (f'tcp:127.0.0.1:{self.monitor.port},'
                              'server,nowait,nodelay')]

        # TODO: extract SSH into a separate plugin?
        self.vm.ssh = SSH(self.vm,
                          key=path.fingertip('ssh_key', 'fingertip.paramiko'))
        self.vm.shared_directory = SharedDirectory(self.vm)
        self.vm.exec = self.vm.ssh.exec
        ssh_host_forward = f'hostfwd=tcp:127.0.0.1:{self.vm.ssh.port}-:22'
        cache_guest_forward = (CACHE_INTERNAL_IP, CACHE_INTERNAL_PORT,
                               f'nc 127.0.0.1 {self.vm.http_cache.port}')
        guest_forwards = guest_forwards + [cache_guest_forward]
        run_args += ['-device', 'virtio-net,netdev=net0', '-netdev',
                     ','.join(['user', 'id=net0', ssh_host_forward] +
                              (['restrict=yes'] if self.vm.sealed else []) +
                              [f'guestfwd=tcp:{ip}:{port}-cmd:{cmd}'
                               for ip, port, cmd in guest_forwards])]

        image = os.path.join(self.vm.path, 'image.qcow2')
        if self._image_to_clone:
            required_space = os.path.getsize(self._image_to_clone) + 2**30
            lock = fasteners.process_lock.InterProcessLock('/tmp/.fingertip')
            lock.acquire()
            if self.vm._transient and temp.has_space(required_space):
                image = temp.disappearing_file('/tmp', hint='fingertip-qemu')
                reflink.auto(self._image_to_clone, image)
                lock.release()
            else:
                lock.release()
                reflink.auto(self._image_to_clone, image)
            self._image_to_clone = None
        run_args += ['-drive',
                     f'file={image},cache=unsafe,if=virtio,discard=unmap']

        run_args += ['-m', self.ram_size]

        os.makedirs(path.SHARED, exist_ok=True)

        args = QEMU_COMMON_ARGS + self.custom_args + run_args + extra_args
        self.vm.log.debug(' '.join(args))
        if self.vm._backend_mode == 'pexpect':
            pexp = self.vm.log.pseudofile_powered(pexpect.spawn,
                                                  logfile=logging.INFO)
            self.vm.console = pexp(self._qemu, args, echo=False,
                                   timeout=None,
                                   encoding='utf-8', codec_errors='ignore')
            self.live = True
        elif self.vm._backend_mode == 'direct':
            subprocess.run([self._qemu, '-serial', 'mon:stdio'] + args,
                           check=True)
            self.live = False
            self._go_down()

    def wait(self):
        self.vm.console.expect(pexpect.EOF)
        self.vm.console.close()
        assert self.vm.console.exitstatus == 0, 'QEMU terminated with an error'
        self._go_down()
        self.live = False

    def _go_down(self):
        if self.live:
            self.vm.console = None
            self.vm.ssh.invalidate()
            del self.vm.exec

    def compress_image(self):
        assert not self.live
        image = os.path.join(self.vm.path, 'image.qcow2')
        self.vm.log.info(f'compressing {image}')
        run = self.vm.log.pipe_powered(subprocess.run, stdout=logging.INFO,
                                       stderr=logging.ERROR)
        run(['qemu-img', 'convert', '-c', '-Oqcow2', image, image + '-tmp'],
            check=True)
        os.rename(image + '-tmp', image)


class VMException(RuntimeError):
    pass


class UnknownVMException(VMException):
    pass


class NotEnoughSpaceForSnapshotException(VMException):
    pass


class Monitor:
    def __init__(self, vm, port=None):
        self.vm = vm
        self.port = port or free_port.find()
        self.vm.log.debug(f'monitor port {self.port}')
        self._sock = None

    def _connect(self, retries=12, timeout=1/32):
        if self._sock is None:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            repeatedly.keep_trying(
                lambda: self._sock.connect(('127.0.0.1', self.port)),
                ConnectionRefusedError, retries=retries, timeout=timeout
            )
            self.vm.log.debug(f'negotiation')
            server_greeting = self._recv()
            assert 'QMP' in server_greeting
            self._execute('qmp_capabilities')
            self._expect({'return': {}})

    def _disconnect(self):
        self._sock.close()
        self._sock = None

    def _send(self, dictionary):
        self._sock.send(json.dumps(dictionary).encode())

    def _recv(self):
        r = b''
        while True:
            r += self._sock.recv(1)
            if not r or b'\n' in r:
                break
        return json.loads(r.decode())

    def _execute(self, cmd, retries=12, timeout=1/32, **kwargs):
        self._connect(retries, timeout)
        self.vm.log.debug(f'executing: {cmd} {kwargs}')
        if not kwargs:
            self._send({'execute': cmd})
        else:
            self._send({'execute': cmd, 'arguments': kwargs})

    def _execute_human_command(self, cmd):
        self._execute(f'human-monitor-command', **{'command-line': cmd})
        r = self._expect(None)
        while 'timestamp' in r and 'event' in r:
            r = self._expect(None)  # ignore these for now
        assert set(r.keys()) == {'return'}
        assert isinstance(r['return'], str)
        r = r['return']
        if 'Error while writing VM state: No space left on device' in r:
            raise NotEnoughSpaceForSnapshotException(r)
        elif r:
            self.vm.log.error(r)
            raise UnknownVMException(r)
        return r

    def _expect(self, what=None):
        reply = self._recv()
        self.vm.log.debug(f'expecting: {what}')
        self.vm.log.debug(f'received: {reply}')
        if what is not None:
            assert reply == what
        return reply

    def resume(self):
        self._execute('cont')
        r = self._expect(None)
        assert set(r.keys()) == {'timestamp', 'event'}
        assert r['event'] == 'RESUME'
        self._expect({'return': {}})

    def pause(self):
        self.vm.hooks.disrupt()
        self._execute('stop')
        r = self._expect(None)
        assert set(r.keys()) == {'timestamp', 'event'}
        assert r['event'] == 'STOP'
        self._expect({'return': {}})

    def quit(self):
        self.vm.hooks.disrupt()
        self._execute('quit')
        self._expect({'return': {}})
        self._disconnect()

    def checkpoint(self, name='tip'):
        self._execute_human_command(f'savevm {name}')

    def restore(self, name='tip'):
        self._execute_human_command(f'loadvm {name}')

    def del_checkpoint(self, name):
        self._execute_human_command(f'delvm {name}')

    def commit(self):
        # saves the changes to current file only,
        # as live QEMU images are not CoW-backed with qcow2 mechanism,
        # just deduplicated with FS-level reflinks that QEMU is unaware of
        self._execute_human_command('commit all')


class SharedDirectory:
    def __init__(self, m):
        self.m = m
        self.mount_count = 0

    def __enter__(self):
        if not self.mount_count:
            self.m('mkdir -p /shared && '
                   'mount -t 9p -o trans=virtio shared /shared '
                   '  -oversion=9p2000.L')
        self.mount_count += 1

    def __exit__(self, *_):
        self.mount_count -= 1
        if not self.mount_count:
            self.m('umount /shared')


class SSH:
    def __init__(self, m, key, host='127.0.0.1', port=None):
        self.host, self.key = host, key
        self.port = port or free_port.find()
        self.m = m
        self.m.log.debug(f'ssh port {self.port}')
        self._transport = None

    def connect(self, force_reconnect=False, retries=12, timeout=1/32):
        import paramiko  # ... in parallel with VM spin-up
        if not force_reconnect and self._transport is not None:
            self._transport.send_ignore()
            if self._transport.is_authenticated():
                return  # the transport is already OK
        self._transport = None
        self.m.log.debug('waiting for the VM to spin up and offer SSH...')
        pkey = paramiko.ECDSAKey.from_private_key_file(self.key)

        def connect():
            self.m.log.debug('Trying to connect ...')
            t = paramiko.Transport((self.host, self.port))
            t.start_client()
            return t

        transport = repeatedly.keep_trying(
            connect,
            paramiko.ssh_exception.SSHException,
            retries=retries, timeout=timeout
        )
        transport.auth_publickey('root', pkey)
        self._transport = transport

    def invalidate(self):
        # gracefully terminate transport channel
        if self._transport:
            self.m.log.debug('Closing SSH session')
            self._transport.close()
        self._transport = None

    def _stream_out_and_err(self, channel):
        sel = selectors.DefaultSelector()
        sel.register(channel, selectors.EVENT_READ)
        out, err, out_buf, err_buf = b'', b'', b'', b''
        while True:
            sel.select()
            if channel.recv_ready():
                r = channel.recv(512)
                out += r
                out_buf += r
                out_lines = out_buf.split(b'\n')
                for out_line in out_lines[:-1]:
                    self.m.log.info(log.strip_control_sequences(out_line))
                out_buf = out_lines[-1]
            elif channel.recv_stderr_ready():
                r = channel.recv_stderr(512)
                err += r
                err_buf += r
                err_lines = err_buf.split(b'\n')
                for err_line in err_lines[:-1]:
                    self.m.log.info(log.strip_control_sequences(err_line))
                err_buf = err_lines[-1]
            else:
                return out, err

    def exec(self, *cmd, shell=False):
        cmd = (' '.join(["'" + a.replace("'", r"\'") + "'" for a in cmd])
               if not shell else cmd[0])
        self.connect()
        try:
            channel = self._transport.open_session()
        except EOFError:
            self.m.log.warning('EOFError on SSH exec, retrying in 2 sec...')
            time.sleep(2)
            self.connect(force_reconnect=True)
            channel = self._transport.open_session()
        self.m.log.info(f'ssh command: {cmd}')
        channel.exec_command(cmd)
        out, err = self._stream_out_and_err(channel)
        retval = channel.recv_exit_status()
        return fingertip.exec.ExecResult(retval, out, err)

    def upload(self, src, dst=None):  # may be unused
        import paramiko  # ... in parallel with VM spin-up
        assert os.path.isfile(src)
        dst = dst or os.path.basename(src)
        self.connect()
        sftp_client = paramiko.SFTPClient.from_transport(self._transport)
        sftp_client.put(src, dst)
        sftp_client.chmod(dst, os.stat(src).st_mode)

    def download(self, src, dst=None):
        import paramiko  # ... in parallel with VM spin-up
        dst = dst or os.path.basename(src)
        self.connect()
        sftp_client = paramiko.SFTPClient.from_transport(self._transport)
        sftp_client.stat(src)  # don't truncate dst if src does not exist
        sftp_client.get(src, dst)

    @property
    def key_file(self):
        key_file = path.fingertip('ssh_key', 'fingertip')
        s = os.stat(key_file)
        mode = s[stat.ST_MODE]
        owner = s[stat.ST_UID]
        # OpenSSH cares about permissions on key file only if the owner
        # matches current user
        if mode & 0o77 and owner == os.getuid():
            self.m.log.debug(f'fixing up permissions on {key_file}')
            os.chmod(key_file, mode & 0o7700)
        return key_file



import subprocess

import fingertip


@fingertip.transient(when='last')
def main(m, no_unseal=False):
    m = m if no_unseal else m.apply('unseal')
    with m:
        m.expiration.cap(0)  # non-deterministic user input, never reuse
        m.log.info(f'waiting for the SSH server to be up...')
        m.ssh.exec('true')
        # terminate the ssh session not to leave any traces in vm
        m.ssh.invalidate()

        m.log.info(f'starting interactive SSH session, {m.ssh.port}')
        m.log.plain()
        subprocess.run(['ssh',
                        '-o', 'StrictHostKeyChecking=no',
                        '-o', 'UserKnownHostsFile=/dev/null',
                        '-o', 'GSSAPIAuthentication=no',
                        '-o', 'GSSAPIKeyExchange=no',
                        '-i', m.ssh.key_file,
                        '-p', str(m.ssh.port),
                        '-t',
                        'root@127.0.0.1'], check=True)
    return m



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.

import fingertip


def make_greeting(m, greeting='Hello!'):
    with m:
        if hasattr(m, 'console'):
            m.console.sendline(f"echo '{greeting}' > .greeting")
            m.console.expect_exact(m.prompt)
            m.console.sendline("ls -l .greeting")
            m.console.expect_exact(m.prompt)
        else:
            raise NotImplementedError()
        return m


@fingertip.transient
def main(m, greeting='Hello!'):
    with m.apply(make_greeting, greeting=greeting) as m:
        assert m('cat .greeting').out.strip() == greeting



# Licensed under GNU General Public License v3 or later, see COPYING.
# Copyright (c) 2019 Red Hat, Inc., see CONTRIBUTORS.


def main(m):
    if m.sealed:
        m.expiration.cap('4h')
        m.sealed = False
        with m:
            m.hooks.unseal()
    return m
